@@ -1,7 +1,7 @@
 //! Opcode definitions. Implemented on Writer/VM.
 
 use std::mem::size_of;
-use crate::{StackAddress, StackOffset};
+use crate::{StackAddress, StackOffset, STACK_ADDRESS_TYPE, RustFnIndex};
 use crate::bytecode::{ARG1, ARG2, ARG3, HeapRef, runtime::{stack::{StackOp, StackRelativeOp}, heap::{HeapOp, HeapCmp, HeapRefOp}, vm::{VMState, CopyTarget}}};
 
 type Data8 = u8;
@@ -9,7 +9,7 @@ type Data16 = u16;
 type Data32 = u32;
 type Data64 = u64;
 
-impl_vm!{
+impl_opcodes!{
 
     /// Does nothing.
     fn noop(&mut self) { }
@@ -29,57 +29,44 @@ impl_vm!{
         self.stack.truncate(self.stack.sp() - size_of::<T>() as StackAddress);
     }
 
-    /// Pushes 8 bit 0 onto stack.
-    fn zero8(&mut self) {
-        self.stack.push(0u8);
-    }
-    /// Pushes 32 bit 0 onto stack.
-    fn zero32(&mut self) {
-        self.stack.push(0u32);
-    }
-    /// Pushes 64 bit 0 onto stack.
-    fn zero64(&mut self) {
-        self.stack.push(0u64);
+    /// Pushes 0 onto stack.
+    fn <
+        zero8<T: Data8>(),
+        zero16<T: Data16>(),
+        zero32<T: Data32>(),
+        zero64<T: Data64>(),
+    >(&mut self) {
+        self.stack.push(0 as T);
     }
 
-    /// Pushes 8 bit 1 onto stack.
-    fn one8(&mut self) {
-        self.stack.push(1u8);
-    }
-    /// Pushes 32 bit 1 onto stack.
-    fn one32(&mut self) {
-        self.stack.push(1u32);
-    }
-    /// Pushes 64 bit 1 onto stack.
-    fn one64(&mut self) {
-        self.stack.push(1u64);
+    /// Pushes 1 onto stack.
+    fn <
+        one8<T: Data8>(),
+        one16<T: Data16>(),
+        one32<T: Data32>(),
+        one64<T: Data64>(),
+    >(&mut self) {
+        self.stack.push(1 as T);
     }
 
-    /// Pushes 8 bit -1 onto stack.
-    fn fill8(&mut self) {
-        self.stack.push(-1i8);
-    }
-    /// Pushes 32 bit -1 onto stack.
-    fn fill32(&mut self) {
-        self.stack.push(-1i32);
-    }
-    /// Pushes 64 bit -1 onto stack.
-    fn fill64(&mut self) {
-        self.stack.push(-1i64);
+    /// Pushes -1 onto stack.
+    fn <
+        fill8<T: i8>(),
+        fill16<T: i16>(),
+        fill32<T: i32>(),
+        fill64<T: i64>(),
+    >(&mut self) {
+        self.stack.push(-1 as T);
     }
 
-    /// Pushes 8 bit value onto stack.
-    fn literali8(&mut self, value: u8) {
+    /// Pushes 8 bit value as an 8 or 32 bit value onto stack.
+    fn <
+        literali8(value: u8),
+        literalu32(value: u8 as u32),
+        literals32(value: i8 as i32),
+    >(&mut self) {
         self.stack.push(value);
     }
-    /// Pushes 32 bit unsigned value onto stack.
-    fn literalu32(&mut self, value: u8) {
-        self.stack.push(value as u32);
-    }
-    /// Pushes 32 bit signed value onto stack (sign extended).
-    fn literals32(&mut self, value: i8) {
-        self.stack.push(value as i32);
-    }
 
     /// Loads 32 bit function argument 1 and pushes it onto the stack. Equals load32(ARG1).
     fn load_arg1(&mut self) {
@@ -99,18 +86,15 @@ impl_vm!{
 
     /// Loads data from constpool at given offset and pushes it onto the stack.
     fn <
-        const8_8<T: Data8>(offset: u8 as StackAddress),
-        const8_16<T: Data8>(offset: u16 as StackAddress),
-        const8_32<T: Data8>(offset: StackAddress),
         const16_8<T: Data16>(offset: u8 as StackAddress),
         const16_16<T: Data16>(offset: u16 as StackAddress),
-        const16_32<T: Data16>(offset: StackAddress),
+        const16_sa<T: Data16>(offset: StackAddress),
         const32_8<T: Data32>(offset: u8 as StackAddress),
         const32_16<T: Data32>(offset: u16 as StackAddress),
-        const32_32<T: Data32>(offset: StackAddress),
+        const32_sa<T: Data32>(offset: StackAddress),
         const64_8<T: Data64>(offset: u8 as StackAddress),
         const64_16<T: Data64>(offset: u16 as StackAddress),
-        const64_32<T: Data64>(offset: StackAddress),
+        const64_sa<T: Data64>(offset: StackAddress),
     >(&mut self) {
         let local: T = self.stack.load(offset);
         self.stack.push(local);
@@ -120,16 +104,16 @@ impl_vm!{
     fn <
         load8_8<T: Data8>(offset: i8 as StackOffset),
         load8_16<T: Data8>(offset: i16 as StackOffset),
-        load8_32<T: Data8>(offset: StackOffset),
+        load8_sa<T: Data8>(offset: StackOffset),
         load16_8<T: Data16>(offset: i8 as StackOffset),
         load16_16<T: Data16>(offset: i16 as StackOffset),
-        load16_32<T: Data16>(offset: StackOffset),
+        load16_sa<T: Data16>(offset: StackOffset),
         load32_8<T: Data32>(offset: i8 as StackOffset),
         load32_16<T: Data32>(offset: i16 as StackOffset),
-        load32_32<T: Data32>(offset: StackOffset),
+        load32_sa<T: Data32>(offset: StackOffset),
         load64_8<T: Data64>(offset: i8 as StackOffset),
         load64_16<T: Data64>(offset: i16 as StackOffset),
-        load64_32<T: Data64>(offset: StackOffset),
+        load64_sa<T: Data64>(offset: StackOffset),
     >(&mut self) {
         let abs = (offset + if offset >= 0 { self.stack.fp as StackOffset } else { self.stack.sp() as StackOffset }) as StackAddress; // fp+offset or sp-offset
         let local: T = self.stack.load(abs);
@@ -140,31 +124,50 @@ impl_vm!{
     fn <
         store8_8<T: Data8>(offset: i8 as StackOffset),
         store8_16<T: Data8>(offset: i16 as StackOffset),
-        store8_32<T: Data8>(offset: StackOffset),
+        store8_sa<T: Data8>(offset: StackOffset),
         store16_8<T: Data16>(offset: i8 as StackOffset),
         store16_16<T: Data16>(offset: i16 as StackOffset),
-        store16_32<T: Data16>(offset: StackOffset),
+        store16_sa<T: Data16>(offset: StackOffset),
         store32_8<T: Data32>(offset: i8 as StackOffset),
         store32_16<T: Data32>(offset: i16 as StackOffset),
-        store32_32<T: Data32>(offset: StackOffset),
+        store32_sa<T: Data32>(offset: StackOffset),
         store64_8<T: Data64>(offset: i8 as StackOffset),
         store64_16<T: Data64>(offset: i16 as StackOffset),
-        store64_32<T: Data64>(offset: StackOffset),
+        store64_sa<T: Data64>(offset: StackOffset),
     >(&mut self) {
         let abs = (offset + if offset >= 0 { self.stack.fp as StackOffset } else { self.stack.sp() as StackOffset }) as StackAddress; // fp+offset or sp-offset
         let local: T = self.stack.pop();
         self.stack.store(abs, local);
     }
 
-    /// Reads value from the n-th stack element relative to the top of the stack and pushes it.
-    /// n=0 is the topmost stack value, n=sizeof(value) the previous value.
+    /// Pops HeapRef off the stack and stores it at the given offset (relative to the stack frame).
+    /// Increases refcount of the new value.
+    fn storex_new(&mut self, index: StackOffset, constructor: StackAddress) {
+        let value: HeapRef = self.stack.pop();
+        self.stack.store_fp(index, value);
+        self.refcount_value(value, constructor, HeapRefOp::Inc);
+    }
+
+    /// Pops HeapRef off the stack and stores it at the given offset (relative to the stack frame).
+    /// Decreses refcount of the previous contents and increases refcount of the new value.
+    fn storex_replace(&mut self, index: StackOffset, constructor: StackAddress) {
+        let prev: HeapRef = self.stack.load_fp(index);
+        let next: HeapRef = self.stack.pop();
+        self.stack.store_fp(index, next);
+        if next != prev {
+            self.refcount_value(next, constructor, HeapRefOp::Inc);
+            self.refcount_value(prev, constructor, HeapRefOp::Dec);
+        }
+    }
+
+    /// Clones the topmost stack value.
     fn <
-        clone8<T: Data8>(n: u8),
-        clone16<T: Data16>(n: u8),
-        clone32<T: Data32>(n: u8),
-        clone64<T: Data64>(n: u8),
+        clone8<T: Data8>(),
+        clone16<T: Data16>(),
+        clone32<T: Data32>(),
+        clone64<T: Data64>(),
     >(&mut self) {
-        let data: T = self.stack.load_sp(- (size_of::<T>() as StackOffset) - n as StackOffset);
+        let data: T = self.stack.load_sp(- (size_of::<T>() as StackOffset));
         self.stack.push(data);
     }
 
@@ -183,7 +186,43 @@ impl_vm!{
         self.stack.store(pos_b, a);
     }
 
-    /// Pop u32 "index" and heap reference and push the resulting heap reference with offset += index * element_size onto the stack.
+    /// Decrements the value at the top of the stack.
+    fn <
+        deci8<T: i8>(decr: i8),
+        deci16<T: i16>(decr: i8),
+        deci32<T: i32>(decr: i8),
+        deci64<T: i64>(decr: i8)
+    >(&mut self) {
+        let a: T = self.stack.pop();
+        self.stack.push(T::wrapping_sub(a, decr as T));
+    }
+
+    /// Decrements the stackvalue at given offset (relative to the stack frame) and pushes the result onto the stack.
+    fn <
+        predeci8<T: i8>(offset: StackOffset, decr: i8),
+        predeci16<T: i16>(offset: StackOffset, decr: i8),
+        predeci32<T: i32>(offset: StackOffset, decr: i8),
+        predeci64<T: i64>(offset: StackOffset, decr: i8)
+    >(&mut self) {
+        let mut value: T = self.stack.load_fp(offset);
+        value = T::wrapping_sub(value, decr as T);
+        self.stack.store_fp(offset, value);
+        self.stack.push(value);
+    }
+
+    /// Decrements the stackvalue at given offset (relative to the stack frame) and pushes the previous value onto the stack.
+    fn <
+        postdeci8<T: i8>(offset: StackOffset, decr: i8),
+        postdeci16<T: i16>(offset: StackOffset, decr: i8),
+        postdeci32<T: i32>(offset: StackOffset, decr: i8),
+        postdeci64<T: i64>(offset: StackOffset, decr: i8)
+    >(&mut self) {
+        let value: T = self.stack.load_fp(offset);
+        self.stack.store_fp(offset, T::wrapping_sub(value, decr as T));
+        self.stack.push(value);
+    }
+
+    /// Pop StackAddress sized "index" and heap reference and push the resulting heap reference with offset += index * element_size onto the stack.
     fn index(&mut self, element_size: u8) {
         let element_index: StackAddress = self.stack.pop();
         let mut item: HeapRef = self.stack.pop();
@@ -191,6 +230,17 @@ impl_vm!{
         self.stack.push(item);
     }
 
+    /// Offsets the heap address at the top of the stack by given value.
+    fn <
+        offsetx_8(offset: i8 as StackOffset),
+        offsetx_16(offset: i16 as StackOffset),
+        offsetx_sa(offset: StackOffset),
+    >(&mut self) {
+        let mut item: HeapRef = self.stack.pop();
+        item.add_offset(offset);
+        self.stack.push(item);
+    }
+
     /// Pops value off the stack, pushes 0 for values < 0 and the original value for values >= 0.
     fn <
         zclampf32<T: f32>(),
@@ -222,14 +272,11 @@ impl_vm!{
         i64_to_f32<F: i64, T: f32>(),
         u64_to_f32<F: u64, T: f32>(),
         f64_to_f32<F: f64, T: f32>(),
-
         i64_to_f64<F: i64, T: f64>(),
         u64_to_f64<F: u64, T: f64>(),
         f32_to_f64<F: f32, T: f64>(),
-
         f32_to_i64<F: f32, T: i64>(),
         f64_to_i64<F: f64, T: i64>(),
-
         f32_to_u64<F: f32, T: u64>(),
         f64_to_u64<F: f64, T: u64>(),
     >(&mut self) {
@@ -441,78 +488,6 @@ impl_vm!{
         self.stack.push(a % b);
     }
 
-    /// Increments the value at the top of the stack.
-    fn <
-        inci8<T: Data8>(),
-        inci16<T: Data16>(),
-        inci32<T: Data32>(),
-        inci64<T: Data64>()
-    >(&mut self) {
-        let a: T = self.stack.pop();
-        self.stack.push(T::wrapping_add(a, 1));
-    }
-
-    /// Decrements the value at the top of the stack.
-    fn <
-        deci8<T: Data8>(),
-        deci16<T: Data16>(),
-        deci32<T: Data32>(),
-        deci64<T: Data64>()
-    >(&mut self) {
-        let a: T = self.stack.pop();
-        self.stack.push(T::wrapping_sub(a, 1));
-    }
-
-    /// Increments the stackvalue at given offset (relative to the stack frame) and pushes the result onto the stack.
-    fn <
-        preinci8<T: Data8>(offset: StackOffset),
-        preinci16<T: Data16>(offset: StackOffset),
-        preinci32<T: Data32>(offset: StackOffset),
-        preinci64<T: Data64>(offset: StackOffset)
-    >(&mut self) {
-        let mut local: T = self.stack.load_fp(offset);
-        local = T::wrapping_add(local, 1);
-        self.stack.store_fp(offset, local);
-        self.stack.push(local);
-    }
-
-    /// Decrements the stackvalue at given offset (relative to the stack frame) and pushes the result onto the stack.
-    fn <
-        predeci8<T: Data8>(offset: StackOffset),
-        predeci16<T: Data16>(offset: StackOffset),
-        predeci32<T: Data32>(offset: StackOffset),
-        predeci64<T: Data64>(offset: StackOffset)
-    >(&mut self) {
-        let mut local: T = self.stack.load_fp(offset);
-        local = T::wrapping_sub(local, 1);
-        self.stack.store_fp(offset, local);
-        self.stack.push(local);
-    }
-
-    /// Increments the stackvalue at given offset (relative to the stack frame) and pushes the previous value onto the stack.
-    fn <
-        postinci8<T: Data8>(offset: StackOffset),
-        postinci16<T: Data16>(offset: StackOffset),
-        postinci32<T: Data32>(offset: StackOffset),
-        postinci64<T: Data64>(offset: StackOffset)
-    >(&mut self) {
-        let local: T = self.stack.load_fp(offset);
-        self.stack.store_fp(offset, T::wrapping_add(local, 1));
-        self.stack.push(local);
-    }
-
-    /// Decrements the stackvalue at given offset (relative to the stack frame) and pushes the previous value onto the stack.
-    fn <
-        postdeci8<T: Data8>(offset: StackOffset),
-        postdeci16<T: Data16>(offset: StackOffset),
-        postdeci32<T: Data32>(offset: StackOffset),
-        postdeci64<T: Data64>(offset: StackOffset)
-    >(&mut self) {
-        let local: T = self.stack.load_fp(offset);
-        self.stack.store_fp(offset, T::wrapping_sub(local, 1));
-        self.stack.push(local);
-    }
-
     /// Pops two values and pushes a 1 if the first value equals the second, otherwise a 0.
     fn <
         ceq8<T: Data8>(),
@@ -592,23 +567,38 @@ impl_vm!{
         }
     }
     /// Jumps to given address if the 8 bit stack-top is 0.
-    fn j0_top(&mut self, addr: StackAddress) {
+    fn j0_nc(&mut self, addr: StackAddress) {
         let a: Data8 = self.stack.top();
         if a == 0 {
             self.pc = addr;
         }
     }
     /// Jumps to given address if the 8 bit stack-top is not 0.
-    fn jn0_top(&mut self, addr: StackAddress) {
+    fn jn0_nc(&mut self, addr: StackAddress) {
         let a: Data8 = self.stack.top();
         if a != 0 {
             self.pc = addr;
         }
     }
 
+    /// Jumps to given address if the StackAddress sized stack-top equals 0.
+    fn j0_sa_nc(&mut self, addr: StackAddress) {
+        let a: StackAddress = self.stack.top();
+        if a == 0 {
+            self.pc = addr;
+        }
+    }
+    /// Jumps to given address if the StackAddress sized stack-top does not equal 0.
+    fn jn0_sa_nc(&mut self, addr: StackAddress) {
+        let a: StackAddress = self.stack.top();
+        if a != 0 {
+            self.pc = addr;
+        }
+    }
+
     /// Constructs an instance of a non-primitive type.
     fn construct(&mut self, constructor: StackAddress, prototype: StackAddress) {
-        let mut constructor = constructor; // impl_vm macro does not allow mut arguments
+        let mut constructor = constructor; // impl_opcodes macro does not allow mut arguments
         let mut prototype = prototype; //if prototype < 0 { self.stack.sp() as i32 + prototype } else { prototype } as u32;
         self.construct_value(&mut constructor, &mut prototype, CopyTarget::Stack, false);
     }
@@ -616,144 +606,196 @@ impl_vm!{
     /// Constructs an instance of a non-primitive type from a prototype that was dynamically constructed on the stack.
     /// This opcode expects any contained strings to already be constructed on the heap and be stored as references on the stack.
     fn construct_dyn(&mut self, constructor: StackAddress, relative_prototype: StackAddress) {
-        let mut constructor = constructor; // impl_vm macro does not allow mut arguments
+        let mut constructor = constructor; // impl_opcodes macro does not allow mut arguments
         let mut prototype = (self.stack.sp() as i32 - relative_prototype as i32) as StackAddress;
         self.construct_value(&mut constructor, &mut prototype, CopyTarget::Stack, true);
     }
 
-    /// Increase reference count for the top heap object on the stack. Does not pop the object off the stack.
+    /// Increase reference count for the top heap reference on the stack. Does not pop the object off the stack.
     fn <
-        cntinc_8(constructor: u8 as StackAddress),
-        cntinc_16(constructor: u16 as StackAddress),
-        cntinc_32(constructor: StackAddress),
+        cntinc_8_nc(constructor: u8 as StackAddress),
+        cntinc_16_nc(constructor: u16 as StackAddress),
+        cntinc_sa_nc(constructor: StackAddress),
     >(&mut self) {
         let item: HeapRef = self.stack.top();
         self.refcount_value(item, constructor, HeapRefOp::Inc);
     }
-    /// Pops a heap object off the stack and decreases its reference count by 1, freeing it on 0.
+    /// Pops a heap reference off the stack and decreases its reference count by 1, freeing it on 0.
     fn <
         cntdec_8(constructor: u8 as StackAddress),
         cntdec_16(constructor: u16 as StackAddress),
-        cntdec_32(constructor: StackAddress),
+        cntdec_sa(constructor: StackAddress),
     >(&mut self) {
         let item: HeapRef = self.stack.pop();
         self.refcount_value(item, constructor, HeapRefOp::Dec);
     }
-    /// Pops a heap object off the stack and decreases its reference count by 1, freeing it on 0.
+    /// Pops a heap reference off the stack and frees it if it has a reference count of 0.
     fn <
-        cntzero_8(constructor: u8 as StackAddress),
-        cntzero_16(constructor: u16 as StackAddress),
-        cntzero_32(constructor: StackAddress),
+        cntfreetmp_8(constructor: u8 as StackAddress),
+        cntfreetmp_16(constructor: u16 as StackAddress),
+        cntfreetmp_sa(constructor: StackAddress),
     >(&mut self) {
-        let item: HeapRef = self.stack.top();
-        self.refcount_value(item, constructor, HeapRefOp::Zero);
+        let item: HeapRef = self.stack.pop();
+        self.refcount_value(item, constructor, HeapRefOp::FreeTmp);
     }
 
-    /// Reads the top value off the stack and pushes it onto the tmp stack.
-    fn cntstore(&mut self) {
-        let value: HeapRef = self.stack.top();
-        self.cnt.push(value);
-    }
-
-    /// Pops the top value off the tmp stack and pushes it onto the stack.
-    fn cntpop(&mut self) {
-        let value: HeapRef = self.cnt.pop();
-        self.stack.push(value);
-    }
 
     /// Calls the given Rust function.
     fn rustcall(&mut self, &mut context, func: RustFn) {
-        T::from_u16(func).exec(self, context);
+        T::from_index(func).exec(self, context);
     }
 
-    /// Function call. Saves state and sets programm counter to given addr. Expects
-    /// callee arguments on the stack.
+    /// Function call. Saves state and sets programm counter to given addr.
     fn call(&mut self, addr: StackAddress, arg_size: StackAddress) {
-        self.stack.push_frame(self.pc);
-        self.stack.fp = self.stack.sp() - arg_size; // new frame starts with arguments
-        self.pc = addr;                             // set new program counter
+        // stack: ... | ARGS
+        self.stack.push(self.stack.fp);
+        self.stack.fp = self.stack.sp() - arg_size - (size_of::<StackAddress>() as StackAddress);
+        self.stack.push(self.pc);
+        self.pc = addr;
+        // stack: ARGS | previous FP | previous PC | (local vars and dynamic stack follow here)
+    }
+
+    /// Function return. Restores state, removes arguments left on stack by caller.
+    fn ret0(&mut self, arg_size: StackAddress) {
+        // stack: ARGS | previous FP | previous PC | local vars
+        let prev_fp = self.stack.load_fp(arg_size as StackOffset);
+        let prev_pc = self.stack.load_fp(arg_size as StackOffset + size_of::<StackAddress>() as StackOffset);
+        self.stack.truncate(self.stack.fp);
+        self.pc = prev_pc;
+        self.stack.fp = prev_fp;
     }
 
     /// Function return. Restores state, removes arguments left on stack by caller and
     /// leaves call result on the stack.
-    fn ret(&mut self, ret_size: u8) {
-        let ret_size = ret_size as StackAddress;
-        // move function result to the beginning of this stack frame
-        if ret_size == 4 {
-            let arg: u32 = self.stack.load(self.stack.sp() - 4);
-            self.stack.store(self.stack.fp, arg);
-        } else if ret_size > 0 {
-            self.stack.copy(self.stack.sp() - ret_size, self.stack.fp, ret_size);
-        }
-        // truncate stack down, so that the result is the last item on the stack
-        self.stack.truncate(self.stack.fp + ret_size);
-        // remove stack frame, restore program counter
-        self.pc = self.stack.pop_frame();
+    fn <
+        ret8<T: Data8>(arg_size: StackAddress),
+        ret16<T: Data16>(arg_size: StackAddress),
+        ret32<T: Data32>(arg_size: StackAddress),
+        ret64<T: Data64>(arg_size: StackAddress),
+    >(&mut self) {
+        // stack: ARGS | previous FP | previous PC | local vars | RESULT
+        let prev_fp = self.stack.load_fp(arg_size as StackOffset);
+        let prev_pc = self.stack.load_fp(arg_size as StackOffset + size_of::<StackAddress>() as StackOffset);
+        let ret: T = self.stack.top();
+        self.stack.store_fp(0, ret);
+        self.stack.truncate(self.stack.fp + size_of::<T>() as StackAddress);
+        self.pc = prev_pc;
+        self.stack.fp = prev_fp;
     }
 
-    /// Pops 2 heap objects dest and src and copies num_bytes bytes from src to dest.
+    /// Pops 2 heap references dest and src and copies num_bytes bytes from src to dest.
     fn heap_copy(&mut self, num_bytes: StackAddress) {
         let dest: HeapRef = self.stack.pop();
         let src: HeapRef = self.stack.pop();
         self.heap.copy(dest, src, num_bytes);
     }
 
-    /// Pops 2 heap objects and compares num_bytes bytes.
+    /// Pops 2 heap references and compares num_bytes bytes.
     fn heap_ceq(&mut self, num_bytes: StackAddress) {
         let b: HeapRef = self.stack.pop();
         let a: HeapRef = self.stack.pop();
         let equals = self.heap.compare(a, b, num_bytes, HeapCmp::Eq);
         self.stack.push(equals as Data8);
+
+        self.heap.ref_item(a.index(), HeapRefOp::FreeTmp);
+        self.heap.ref_item(b.index(), HeapRefOp::FreeTmp);
     }
-    /// Pops 2 heap objects and compares num_bytes bytes.
+    /// Pops 2 heap references and compares num_bytes bytes.
     fn heap_cneq(&mut self, num_bytes: StackAddress) {
         let b: HeapRef = self.stack.pop();
         let a: HeapRef = self.stack.pop();
         let equals = self.heap.compare(a, b, num_bytes, HeapCmp::Neq);
         self.stack.push(equals as Data8);
+
+        self.heap.ref_item(a.index(), HeapRefOp::FreeTmp);
+        self.heap.ref_item(b.index(), HeapRefOp::FreeTmp);
     }
 
-    /// Pops 2 heap string slices, compares them for equality and pushes the result.
+    /// Pops 2 heap references to strings, compares the strings for equality and pushes the result.
     fn string_ceq(&mut self) {
         let b: HeapRef = self.stack.pop();
         let a: HeapRef = self.stack.pop();
         let equals = self.heap.compare_string(a, b, HeapCmp::Eq);
         self.stack.push(equals as Data8);
+
+        self.heap.ref_item(a.index(), HeapRefOp::FreeTmp);
+        self.heap.ref_item(b.index(), HeapRefOp::FreeTmp);
     }
-    /// Pops 2 heap string slices, compares them for inequality and pushes the result.
+    /// Pops 2 heap references to strings, compares the strings for inequality and pushes the result.
     fn string_cneq(&mut self) {
         let b: HeapRef = self.stack.pop();
         let a: HeapRef = self.stack.pop();
         let equals = self.heap.compare_string(a, b, HeapCmp::Neq);
         self.stack.push(equals as Data8);
+
+        self.heap.ref_item(a.index(), HeapRefOp::FreeTmp);
+        self.heap.ref_item(b.index(), HeapRefOp::FreeTmp);
     }
-    /// Pops 2 heap string slices, compares them lexicographically and pushes the result.
+    /// Pops 2 heap references to strings, compares the strings lexicographically and pushes the result.
     fn string_clt(&mut self) {
         let b: HeapRef = self.stack.pop();
         let a: HeapRef = self.stack.pop();
         let equals = self.heap.compare_string(a, b, HeapCmp::Lt);
         self.stack.push(equals as Data8);
+
+        self.heap.ref_item(a.index(), HeapRefOp::FreeTmp);
+        self.heap.ref_item(b.index(), HeapRefOp::FreeTmp);
     }
-    /// Pops 2 heap string slices, compares them lexicographically and pushes the result.
+    /// Pops 2 heap references to strings, compares the strings lexicographically and pushes the result.
     fn string_clte(&mut self) {
         let b: HeapRef = self.stack.pop();
         let a: HeapRef = self.stack.pop();
         let equals = self.heap.compare_string(a, b, HeapCmp::Lte);
         self.stack.push(equals as Data8);
+
+        self.heap.ref_item(a.index(), HeapRefOp::FreeTmp);
+        self.heap.ref_item(b.index(), HeapRefOp::FreeTmp);
     }
-    /// Pops two heap slices, concatenates the heap objects into a new object and pushes its heap slice.
-    fn string_concat(&mut self) {
-        let src_b: HeapRef = self.stack.pop();
-        let src_b_len = self.heap.size_of(src_b.index());
-        let src_a: HeapRef = self.stack.pop();
-        let src_a_len = self.heap.size_of(src_a.index());
+    /// Pops two heap references to strings, concatenates the referenced strings into a new object and pushes its heap reference.
+    fn string_concatx(&mut self) {
+        let b: HeapRef = self.stack.pop();
+        let b_len = self.heap.size_of(b.index());
+        let a: HeapRef = self.stack.pop();
+        let a_len = self.heap.size_of(a.index());
+
         let dest_index: StackAddress = self.heap.alloc(Vec::new());
-        self.heap.copy(HeapRef::new(dest_index, 0), src_a, src_a_len);
-        self.heap.copy(HeapRef::new(dest_index, src_a_len), src_b, src_b_len);
+        self.heap.copy(HeapRef::new(dest_index, 0), a, a_len);
+        self.heap.copy(HeapRef::new(dest_index, a_len), b, b_len);
         self.stack.push(HeapRef::new(dest_index, 0));
+
+        self.heap.ref_item(a.index(), HeapRefOp::FreeTmp);
+        self.heap.ref_item(b.index(), HeapRefOp::FreeTmp);
     }
 
-    /// Pop a heap object and push the heap value at its current offset onto the stack.
+    /// Decrements the stackvalue at given offset (relative to the stack frame) and pushes the result onto the stack.
+    fn <
+        heap_predeci8<T: i8>(decr: i8),
+        heap_predeci16<T: i16>(decr: i8),
+        heap_predeci32<T: i32>(decr: i8),
+        heap_predeci64<T: i64>(decr: i8)
+    >(&mut self) {
+        let item: HeapRef = self.stack.pop();
+        let mut value: T = self.heap.read(item);
+        value = T::wrapping_sub(value, decr as T);
+        self.heap.write(item, value);
+        self.stack.push(value); // push after inc/dec
+    }
+
+    /// Decrements the stackvalue at given offset (relative to the stack frame) and pushes the result onto the stack.
+    fn <
+        heap_postdeci8<T: i8>(decr: i8),
+        heap_postdeci16<T: i16>(decr: i8),
+        heap_postdeci32<T: i32>(decr: i8),
+        heap_postdeci64<T: i64>(decr: i8)
+    >(&mut self) {
+        let item: HeapRef = self.stack.pop();
+        let mut value: T = self.heap.read(item);
+        self.stack.push(value); // push before inc/dec
+        value = T::wrapping_sub(value, decr as T);
+        self.heap.write(item, value);
+    }
+
+    /// Pop a heap reference and push the heap value at its current offset onto the stack.
     fn <
         heap_fetch8<T: Data8>(),
         heap_fetch16<T: Data16>(),
@@ -765,19 +807,41 @@ impl_vm!{
         self.stack.push(data);
     }
 
-    /// Pop a heap object and a value and store the value at current offset in the heap.
+    /// Pop a value and a heap reference and store the value at current offset in the heap.
     fn <
         heap_put8<T: Data8>(),
         heap_put16<T: Data16>(),
         heap_put32<T: Data32>(),
         heap_put64<T: Data64>(),
     >(&mut self) {
-        let item: HeapRef = self.stack.pop();
         let value: T = self.stack.pop();
+        let item: HeapRef = self.stack.pop();
         self.heap.write(item, value);
     }
 
-    /// Pop a heap object and push the heap value at its current offset + given offset onto the stack.
+    /// Pop a value and a heap reference and store the value at current offset in the heap.
+    /// Increases refcount of the new value.
+    fn heap_putx_new(&mut self, constructor: StackAddress) {
+        let value: HeapRef = self.stack.pop();
+        let item: HeapRef = self.stack.pop();
+        self.heap.write(item, value);
+        self.refcount_value(value, constructor, HeapRefOp::Inc);
+    }
+
+    /// Pop a value and a heap reference and store the value at current offset in the heap.
+    /// Decreses refcount of the previous contents and increases refcount of the new value.
+    fn heap_putx_replace(&mut self, constructor: StackAddress) {
+        let next: HeapRef = self.stack.pop();
+        let item: HeapRef = self.stack.pop();
+        let prev: HeapRef = self.heap.read(item);
+        self.heap.write(item, next);
+        if next != prev {
+            self.refcount_value(next, constructor, HeapRefOp::Inc);
+            self.refcount_value(prev, constructor, HeapRefOp::Dec);
+        }
+    }
+
+    /// Pop a heap reference and push the heap value at its current offset + given offset onto the stack.
     fn <
         heap_fetch_member8<T: Data8>(offset: StackAddress),
         heap_fetch_member16<T: Data16>(offset: StackAddress),
@@ -789,7 +853,7 @@ impl_vm!{
         self.stack.push(data);
     }
 
-    /// Pop an element index and heap object and push the heap value at element index onto the stack.
+    /// Pop an element index and heap reference and push the heap value at element index onto the stack.
     fn <
         heap_fetch_element8<T: Data8>(),
         heap_fetch_element16<T: Data16>(),
@@ -802,6 +866,20 @@ impl_vm!{
         self.stack.push(data);
     }
 
+    /// Read an element index (at top) and heap reference (just below) and push the heap value at element index from the end of the heap object onto the stack.
+    fn <
+        heap_tail_element8_nc<T: Data8>(),
+        heap_tail_element16_nc<T: Data16>(),
+        heap_tail_element32_nc<T: Data32>(),
+        heap_tail_element64_nc<T: Data64>(),
+    >(&mut self) {
+        let element_index: StackAddress = self.stack.top();
+        let item: HeapRef = self.stack.load_sp(-((STACK_ADDRESS_TYPE.primitive_size() + HeapRef::primitive_size()) as StackOffset));
+        let offset = self.heap.size_of(item.index()) - size_of::<T>() as StackAddress * (element_index + 1);
+        let data: T = self.heap.read(item.with_offset(offset as StackOffset));
+        self.stack.push(data);
+    }
+
     /// Yield program execution.
     fn yld(&mut self) return {
         self.state = VMState::Yielded;
