@@ -1,6 +1,5 @@
 use std::fmt::{self, Display};
 use crate::frontend::ast::Position;
-use crate::shared::compute_loc;
 
 /// Represents the various possible parser error-kinds.
 #[derive(Copy, Clone, Debug)]
@@ -15,22 +14,33 @@ pub enum ParseErrorKind {
     IllegalStructDef,
     IllegalIfBlock,
     IllegalLetStatement,
+    IllegalModuleDef,
+    DisabledFeature(&'static str),
 }
 
 /// An error reported by the parser (e.g. syntax error).
-#[derive(Copy, Clone, Debug)]
+#[derive(Clone, Debug)]
 pub struct ParseError {
-    pub kind: ParseErrorKind,
-    position: Position, // this is the position from the end of the input
+    kind: ParseErrorKind,
+    position: Position,
+    module_path: String,
 }
 
 impl ParseError {
-    pub(crate) fn new(kind: ParseErrorKind, offset: Position) -> ParseError {
-        Self { kind: kind, position: offset }
+    pub(crate) fn new(kind: ParseErrorKind, position: Position, module_path: &str) -> ParseError {
+        Self { kind, position, module_path: module_path.to_string() }
     }
-    /// Computes and returns the source code location of this error.
-    pub fn loc(self: &Self, input: &str) -> (Position, Position) {
-        compute_loc(input, input.len() as Position - self.position)
+    /// Compute 1-based line/column number from Position (absolute offset from end) in string.
+    pub fn loc(self: &Self, input: &str) -> (u32, u32) {
+        self.position.loc(input)
+    }
+    /// The kind of the error.
+    pub fn kind(self: &Self) -> &ParseErrorKind {
+        &self.kind
+    }
+    /// Path to the module where the error occured.
+    pub fn module_path(self: &Self) -> &str {
+        &self.module_path
     }
 }
 
