@@ -1,20 +1,22 @@
 use crate::shared::typed_ids::TypeId;
+use crate::RustFnIndex;
 
-/// Binding meta information
+/// Binding meta information.
 pub struct BindingInfo {
     pub mutable: bool,
     pub type_id: Option<TypeId>,
 }
 
+/// Function mata information.
 #[derive(Clone)]
 pub struct FunctionInfo {
-    pub ret_type: Option<TypeId>,
-    pub arg_type: Vec<Option<TypeId>>,
     pub kind    : Option<FunctionKind>,
+    pub arg_type: Vec<Option<TypeId>>,
+    pub ret_type: Option<TypeId>,
 }
 
 impl FunctionInfo {
-    pub fn rust_fn_index(self: &Self) -> Option<u16> {
+    pub fn rust_fn_index(self: &Self) -> Option<RustFnIndex> {
         match self.kind {
             Some(FunctionKind::Rust(index)) => Some(index),
             _ => None,
@@ -25,10 +27,12 @@ impl FunctionInfo {
     }
 }
 
+/// The kind of a function described by a FunctionInfo.
 #[derive(Copy, Clone, Debug, PartialEq)]
 pub enum FunctionKind {
-    User,
-    Rust(u16),
+    Function,
+    Method(TypeId),
+    Rust(RustFnIndex),
     Intrinsic(Intrinsic),
 }
 
