@@ -447,6 +447,20 @@ fn branching() {
     assert_all(&result, &[ 1i32, 4, 2 ]);
 }
 
+#[test]
+fn branching_ref_cleanup() {
+    let result = run("
+        if true {
+            let a = \"Hello\";
+            ret_string(a);
+        } else {
+            let b = \"World\";
+            ret_string(b);
+        }
+    ");
+    assert_all(&result, &[ "Hello".to_string() ]);
+}
+
 #[test]
 fn array_literal() {
     let result = run("
@@ -514,6 +528,17 @@ fn array_inference2() {
     assert_all(&result, &[ 1i8, 2, 3 ]);
 }
 
+#[test]
+fn array_mixed_construction() {
+    let result = run("
+        let x = \"Hello\";
+        for w in [ x, \"World\" ] {
+            ret_str(w);
+        }
+    ");
+    assert_all(&result, &[ "Hello".to_string(), "World".to_string() ]);
+}
+
 #[test]
 fn struct_access() {
     let result = run("
@@ -709,37 +734,37 @@ fn index_assign_primitive() {
 
 #[test]
 fn heap_compound_assign() {
-    let result = run("
-        fn left() -> u64 {
+    let result = run(&format!("
+        fn left() -> {:?} {{
             ret_u8(9);
             0
-        }
-        fn right() -> u8 {
+        }}
+        fn right() -> u8 {{
             ret_u8(6);
             1
-        }
-        fn main() {
+        }}
+        fn main() {{
             let test = [ 0u8 ];
             ret_u8(test[0]);
             test[left()] += right();
             ret_u8(test[0]);
-        }
-    ");
-    assert_all(&result, &[ 0u8, 6, 9, 1 ]);
+        }}
+    ", itsy::STACK_ADDRESS_TYPE));
+    assert_all(&result, &[ 0u8, 9, 6, 1 ]);
 }
 
 #[test]
 fn heap_compound_assign64() {
-    let result = run("
-        fn left() -> u64 {
+    let result = run(&format!("
+        fn left() -> {:?} {{
             ret_u64(9);
             0
-        }
-        fn right() -> u64 {
+        }}
+        fn right() -> u64 {{
             ret_u64(6);
             1
-        }
-        fn main() {
+        }}
+        fn main() {{
             let test = [ 0u64 ];
             ret_u64(test[0]);
             test[left()] += right();
@@ -748,9 +773,9 @@ fn heap_compound_assign64() {
             ret_u64(test[0]);
             test[left()] -= right();
             ret_u64(test[0]);
-        }
-    ");
-    assert_all(&result, &[ 0u64, 6, 9, 1, 6, 9, 2, 6, 9, 1 ]);
+        }}
+    ", itsy::STACK_ADDRESS_TYPE));
+    assert_all(&result, &[ 0u64, 9, 6, 1, 9, 6, 2, 9, 6, 1 ]);
 }
 
 #[test]
@@ -900,6 +925,54 @@ fn for_in_inference_assign() {
     assert_all(&result, &[ 2u8, 4, 6, 6 ]);
 }
 
+#[test]
+fn for_in_array() {
+    let result = run("
+        for x in [ 7u8, 11, 13, 17 ] {
+            ret_u8(x);
+        }
+    ");
+    assert_all(&result, &[ 7u8, 11, 13, 17 ]);
+}
+
+#[test]
+fn for_in_array_var() {
+    let result = run("
+        let array = [ 7u8, 11, 13, 17 ];
+        for x in array {
+            ret_u8(x);
+        }
+    ");
+    assert_all(&result, &[ 7u8, 11, 13, 17 ]);
+}
+
+#[test]
+fn for_in_array_struct() {
+    let result = run("
+        struct Test {
+            inner: u8,
+        }
+        fn main() {
+            let array = [ Test { inner: 7 }, Test { inner: 17 }];
+            for x in array {
+                ret_u8(x.inner);
+            }
+        }
+    ");
+    assert_all(&result, &[ 7u8, 17 ]);
+}
+
+#[test]
+fn for_in_array_string() {
+    let result = run("
+        let array = [ \"Hello\", \"World\" ];
+        for x in array {
+            ret_string(x);
+        }
+    ");
+    assert_all(&result, &[ "Hello".to_string(), "World".to_string() ]);
+}
+
 #[test]
 fn recursion() {
     let result = run("
@@ -1244,6 +1317,48 @@ fn string_concat() {
     ]);
 }
 
+#[test]
+fn string_compound_concat() {
+    let result = run("
+        let mut result = \"Hello\";
+        result += \"World\";
+        ret_bool(result == \"HelloWorld\");
+        ret_bool(result != \"Something else\");
+        ret_bool(result != \"HelloWorld\");
+    ");
+    assert_all(&result, &[ true, true, false ]);
+}
+
+#[test]
+fn string_compound_heap_concat() {
+    let result = run("
+        struct Test {
+            s: String
+        }
+        fn main() {
+            let mut result = Test { s: \"Hello\" };
+            result.s += \"World\";
+            ret_bool(result.s == \"HelloWorld\");
+            ret_bool(result.s != \"Something else\");
+            ret_bool(result.s != \"HelloWorld\");
+        }
+    ");
+    assert_all(&result, &[ true, true, false ]);
+}
+
+#[test]
+fn string_loop_concat() {
+    let result = run("
+        let test = \"Hello World\";
+        for i in 0..=5 {
+            test = test + (i as String);
+        }
+        ret_string(test);
+        ret_str(test);
+    ");
+    assert_all(&result, &[ "Hello World012345".to_string(), "Hello World012345".to_string() ]);
+}
+
 #[test]
 fn dead_code_result() {
     let result = run("
